"""Conversion utils."""
import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import yaml

import venomx as vx
from venomx.model.embeddings_pa import pyarrow_schema

logger = logging.getLogger(__name__)


class EmbeddingFormat(str, Enum):
    PARQUET = "parquet"
    YAML = "yaml"
    JSON = "json"


SUFFIX_MAP = {EmbeddingFormat.PARQUET: ".parquet"}


def embeddings_file_tuple(
    path: Union[str, Path],
    embedding_format: Optional[EmbeddingFormat] = EmbeddingFormat.PARQUET,
    allow_bundled=False,
    exists_check=True,
) -> Tuple[Path, Optional[Path], EmbeddingFormat]:
    """
    Given a path, return a tuple of the metadata file and embedded data file.

    :param path:
    :return:
    """
    path = str(path)
    suffix = SUFFIX_MAP.get(embedding_format)
    if path.endswith(".yaml"):
        metadata = Path(path)
        embeddings = Path(path.replace(".yaml", suffix))
    elif path.endswith(suffix):
        metadata = Path(path.replace(suffix, ".yaml"))
        embeddings = Path(path)
    else:
        raise ValueError(f"Invalid path: {path}")
    if exists_check:
        if not metadata.exists():
            raise ValueError(f"Metadata file does not exist: {metadata}")
        if not embeddings.exists():
            if allow_bundled:
                embeddings = None
            else:
                raise ValueError(f"Embeddings file does not exist: {embeddings}")
    return metadata, embeddings, embedding_format


def load_embeddings_as_pandas(
    self, source: Union[str, Path], format: EmbeddingFormat = EmbeddingFormat.PARQUET, **kwargs
) -> pd.DataFrame:
    if format == EmbeddingFormat.PARQUET:
        read_table = pq.read_table(str(source))
        df = read_table.to_pandas()
        df["values"] = df["values"].apply(lambda x: x.tolist() if isinstance(x, np.ndarray) else x)
        # df['values'] = df['values'].apply(lambda x: list(x) if isinstance(x, tuple) else x)
        return df
    else:
        raise ValueError(f"Unsupported format: {format}")


def _is_all_in_one(format: EmbeddingFormat) -> bool:
    """
    Check if a file is an all-in-one file.

    :param format:
    :param source:
    :return:
    """
    return format in [EmbeddingFormat.YAML, EmbeddingFormat.JSON]


def load_index(source: Union[str, Path], format: EmbeddingFormat = None, check=True, **kwargs) -> vx.Index:
    """
    Load an index from a file.

    :param source:
    :param kwargs:
    :return:
    """
    if format is None:
        format = EmbeddingFormat.PARQUET
    all_in_one = _is_all_in_one(format)
    if all_in_one:
        metadata = source
        embeddings = None
        ef = format
    else:
        metadata, embeddings, ef = embeddings_file_tuple(source, format, allow_bundled=all_in_one)
    metadata_obj = yaml.safe_load(open(metadata))
    ix = vx.Index(**metadata_obj)
    if embeddings:
        if format == EmbeddingFormat.PARQUET:
            ix.embeddings_frame = load_embeddings_as_pandas(ix, embeddings, ef, **kwargs)
        else:
            raise NotImplementedError(f"Unsupported format: {format}")
    if not ix.embeddings and ix.embeddings_frame is None:
        raise ValueError(f"Index has no embeddings: {source}")
    if ix.embeddings and not ix.embeddings_frame:
        ix.embeddings_frame = pd.DataFrame([row.model_dump() for row in ix.embeddings])
    if check:
        validate(ix)
    return ix


def validate(ix: vx.Index, **kwargs):
    """
    Validate an index.

    :param ix:
    :param kwargs:
    :return:
    """
    if ix.embeddings_frame is not None:
        vseries = ix.embeddings_frame["values"]
        lens = set(vseries.apply(len).tolist())
        if len(lens) != 1:
            raise ValueError(f"Index has inconsistent embeddings lengths: {lens}")
        vseries_len = list(lens)[0]
        if ix.embeddings_dimensions is None:
            ix.embeddings_dimensions = vseries_len
        if vseries_len != ix.embeddings_dimensions:
            raise ValueError(
                "Index has inconsistent embeddings dimensions: " f"{vseries_len} != {ix.embeddings_dimensions}"
            )


def save_index(ix: vx.Index, target: Union[str, Path], format: EmbeddingFormat = EmbeddingFormat.PARQUET, **kwargs):
    """
    Save an index to a file.

    >>> num_entities = 100
    >>> embedding_dim = 50
    >>> entities = [f"X:{i}" for i in range(num_entities)]
    >>> embeddings = np.random.rand(num_entities, embedding_dim).tolist()
    >>> df = pd.DataFrame({"id": entities, "values": embeddings})
    >>> ix = vx.Index()
    >>> ix.objects = [vx.NamedObject(id=x) for x in entities]
    >>> ix.embeddings_frame = df
    >>> len(ix.objects)
    100
    >>> save_index(ix, "tests/output/test.vx.yaml")
    >>> ix2 = load_index("tests/output/test.vx.yaml")
    >>> len(ix2.objects)
    100

    :param ix:
    :param target:
    :param kwargs:
    :return:
    """
    metadata_path, embeddings_path, ef = embeddings_file_tuple(target, exists_check=False)
    all_in_one = format in [EmbeddingFormat.YAML, EmbeddingFormat.JSON]
    metadata_obj = ix.model_dump(exclude_unset=True)
    if all_in_one:
        metadata_obj["embeddings"] = populate_embeddings_list(ix)
    else:
        metadata_obj["embeddings"] = None
    metadata_obj["embeddings_frame"] = None
    yaml.safe_dump(metadata_obj, open(metadata_path, "w", encoding="utf-8"), sort_keys=False)
    if ix.embeddings_frame is not None:
        if format == EmbeddingFormat.PARQUET:
            df = ix.embeddings_frame
            df["values"] = df["values"].apply(lambda x: pa.array(x, type=pa.float32()))
            schema = pyarrow_schema()
            table = pa.Table.from_pandas(df, schema=schema)
            pq.write_table(table, str(embeddings_path))
        elif all_in_one:
            pass
        else:
            raise ValueError(f"Unsupported format: {format}")


def populate_embeddings_list(ix: vx.Index) -> List[Dict[str, list[float]]]:
    """
    Populate the embeddings list.

    :param ix:
    :return:
    """
    if ix.embeddings_frame is not None:
        embeddings = []
        for _, row in ix.embeddings_frame.iterrows():
            e = vx.Embedding(
                id=row["id"],
                values=row["values"],
            )
            embeddings.append({"id": row["id"], "values": row["values"]})
    return embeddings



from functools import lru_cache

import pyarrow as pa

ID = "id"
NAME = "name"
VALUES = "values"
METADATA = "metadata"


@lru_cache
def pyarrow_schema() -> pa.Schema:
    """
    Schema for the embeddings table.

    :return:
    """
    return pa.schema(
        [
            (ID, pa.string()),
            (VALUES, pa.list_(pa.float32())),
        ]
    )



"""venomx package."""

import importlib_metadata

from venomx.model.venomx import Embedding, Index, NamedObject

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover

__all__ = [
    "Index",
    "Embedding",
    "NamedObject",
]



from __future__ import annotations

import sys
from typing import Any, List, Optional

from pydantic import BaseModel as BaseModel
from pydantic import ConfigDict, Field

if sys.version_info >= (3, 8):
    pass
else:
    pass


metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra="forbid",
        arbitrary_types_allowed=True,
        use_enum_values=True,
    )
    pass


class Index(ConfiguredBaseModel):

    """
    A collection of named objects
    """

    id: Optional[str] = Field(None, description="""The identifier for index""")
    prefixes: Optional[List[Prefix]] = Field(default_factory=list, description="""The prefix map for index""")
    dataset: Optional[Dataset] = Field(None, description="""The dataset for named objects""")
    embeddings: Optional[List[Embedding]] = Field(
        default_factory=list, description="""The embeddings for named objects"""
    )
    embeddings_frame: Optional[Any] = Field(None, description="""Can be a dataframe, pyarrow object...""")
    embeddings_dimensions: Optional[int] = Field(None, description="""The number of dimensions for the embeddings""")
    embedding_model: Optional[Model] = Field(None, description="""The embedding model for dataset""")
    embedding_input_method: Optional[ModelInputMethod] = Field(
        None, description="""The method used to generate the input for the embedding model"""
    )
    objects: Optional[List[NamedObject]] = Field(default_factory=list, description="""The named objects""")
    is_external: Optional[bool] = Field(None, description="""Whether the objects are external""")
    md5: Optional[str] = Field(None, description="""The md5 for the index""")


class Prefix(ConfiguredBaseModel):

    """
    A prefix
    """

    prefix: Optional[str] = Field(None, description="""The prefix for prefix""")
    namespace: Optional[str] = Field(None, description="""The expansion for prefix""")


class Dataset(ConfiguredBaseModel):

    """
    A description of the dataset that the index is over. Note that this is not intended to be a comprehensive description of the dataset (use other standards for this), but enough to give context to the index.
    """

    name: Optional[str] = Field(None, description="""The name for dataset""")
    identifiers: Optional[List[str]] = Field(default_factory=list, description="""Identifiers for the dataset""")
    filesystem_path: Optional[str] = Field(None, description="""The filesystem path for dataset""")
    url: Optional[str] = Field(None, description="""The url for dataset""")
    title: Optional[str] = Field(None, description="""The label for dataset""")
    source: Optional[str] = Field(None, description="""The source for dataset""")
    version: Optional[str] = Field(None, description="""The version for dataset""")
    metadata: Optional[List[str]] = Field(default_factory=list, description="""The metadata for dataset""")


class Model(ConfiguredBaseModel):

    """
    A model
    """

    name: Optional[str] = Field(None, description="""The name for model""")
    identifiers: Optional[List[str]] = Field(default_factory=list, description="""Identifiers for the model""")
    url: Optional[str] = Field(None, description="""The url for model""")
    title: Optional[str] = Field(None, description="""The label for model""")
    source: Optional[str] = Field(None, description="""The source for model""")
    version: Optional[str] = Field(None, description="""The version for model""")
    metadata: Optional[List[str]] = Field(default_factory=list, description="""The metadata for model""")


class ModelInputMethod(ConfiguredBaseModel):

    """
    A method for generating the input for an embedding model. This might be as simple as extracting the label (name) of the object, or it might be involve concatenation of multiple fields.
    """

    fields: Optional[List[str]] = Field(default_factory=list, description="""The fields for model input method""")


class NamedObject(ConfiguredBaseModel):

    """
    A named object
    """

    id: str = Field(..., description="""The identifier for named object""")
    label: Optional[str] = Field(None, description="""The label for named object""")
    metadata: Optional[List[str]] = Field(default_factory=list, description="""The metadata for named object""")


class Embedding(ConfiguredBaseModel):

    id: str = Field(..., description="""The identifier for named object""")
    text: Optional[str] = Field(None, description="""The text for named object""")
    values: Optional[List[float]] = Field(default_factory=list, description="""The embedding for named object""")


class MetadataObject(ConfiguredBaseModel):

    """
    A metadata object
    """

    key: str = Field(...)
    value: Optional[Any] = Field(None)


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
Index.model_rebuild()
Prefix.model_rebuild()
Dataset.model_rebuild()
Model.model_rebuild()
ModelInputMethod.model_rebuild()
NamedObject.model_rebuild()
Embedding.model_rebuild()
MetadataObject.model_rebuild()



"""Command line interface for venomx."""

import logging

import click

__all__ = [
    "main",
]

from venomx.tools.file_io import EmbeddingFormat, load_index, save_index

logger = logging.getLogger(__name__)


input_embeddings_format_option = click.option(
    "-f",
    "--input-embeddings-format",
    type=click.Choice([x.value for x in EmbeddingFormat]),
    default="parquet",
    help="Format of the input embeddings.",
)
output_embeddings_format_option = click.option(
    "-t",
    "--output-embeddings-format",
    type=click.Choice([x.value for x in EmbeddingFormat]),
    default="parquet",
    help="Format of the output embeddings.",
)
output_option = click.option(
    "-o",
    "--output",
    required=True,
    type=click.Path(),
    help="Output path.",
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
def main(verbose: int, quiet: bool):
    """
    CLI for venomx.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)


@main.command()
@input_embeddings_format_option
@click.argument("input_file")
def validate(input_file: str, input_embeddings_format: str):
    """Validate an index."""
    load_index(input_file, format=input_embeddings_format, check=True)


@main.command()
@click.argument("input_file")
@input_embeddings_format_option
@output_embeddings_format_option
@output_option
def convert(input_file: str, input_embeddings_format: str, output_embeddings_format: str, output: str):
    """Merge an index."""
    ix = load_index(input_file, format=EmbeddingFormat(input_embeddings_format), check=True)
    save_index(ix, output, format=EmbeddingFormat(output_embeddings_format))


@main.command()
@click.argument("input_file")
def info(input_file: str):
    """Show information."""
    ix = load_index(input_file, check=True)
    print(f"Num objects: {len(ix.objects)}")


if __name__ == "__main__":
    main()
